"""Error module."""


class ProjectorError(Exception):
    """Base class for errors."""

    def __init__(self, *args, message=None, **_kwargs):
        """Initialize base error."""

        super().__init__(*args, message)


class ProjectorUnavailableError(ProjectorError):
    """Projector unavailable error."""



"""HTTP connection of Epson projector module."""
import logging

import aiohttp
import asyncio
import async_timeout

from .const import (
    ACCEPT_ENCODING,
    ACCEPT_HEADER,
    BUSY,
    EPSON_KEY_COMMANDS,
    DIRECT_SEND,
    HTTP_OK,
    STATE_UNAVAILABLE,
    POWER,
    EPSON_CODES,
    TCP_SERIAL_PORT,
    SERIAL_BYTE,
    JSON_QUERY,
)
from .error import ProjectorUnavailableError
from .timeout import get_timeout

_LOGGER = logging.getLogger(__name__)


class ProjectorHttp:
    """
    Epson projector class.

    Control your projector with Python.
    """

    def __init__(self, host, websession, port=80):
        """
        Epson Projector controller.

        :param str host:        IP address or hostname of Projector
        :param int port:        Port to connect to. Default 80.
        :param bool encryption: User encryption to connect

        """
        self._host = host
        self._http_url = f"http://{self._host}:{port}/cgi-bin/"
        self._headers = {
            "Accept-Encoding": ACCEPT_ENCODING,
            "Accept": ACCEPT_HEADER,
            "Referer": f"http://{self._host}:{port}/cgi-bin/webconf",
        }
        self._serial = None
        self.websession = websession

    def close(self):
        return

    async def get_property(self, command, timeout):
        """Get property state from device."""
        response = await self.send_request(
            timeout=timeout, params=EPSON_KEY_COMMANDS[command], type=JSON_QUERY
        )
        if not response:
            return False
        try:
            if response == STATE_UNAVAILABLE:
                return STATE_UNAVAILABLE
            return response["projector"]["feature"]["reply"]
        except KeyError:
            return BUSY

    async def send_command(self, command, timeout):
        """Send command to Epson."""
        response = await self.send_request(
            timeout=timeout, params=EPSON_KEY_COMMANDS[command], type=DIRECT_SEND
        )
        return response

    async def send_request(self, params, timeout, type=JSON_QUERY):
        """Send request to Epson."""
        try:
            with async_timeout.timeout(timeout):
                url = "{url}{type}".format(url=self._http_url, type=type)
                async with self.websession.get(
                    url=url, params=params, headers=self._headers
                ) as response:
                    if response.status != HTTP_OK:
                        _LOGGER.warning("Error message %d from Epson.", response.status)
                        return False
                    if type == JSON_QUERY:
                        return await response.json()
                    return response
        except (
            aiohttp.ClientError,
            aiohttp.ClientConnectionError,
            TimeoutError,
            asyncio.exceptions.TimeoutError,
        ):
            raise ProjectorUnavailableError(STATE_UNAVAILABLE)

    async def get_serial(self):
        """Send TCP request for serial to Epson."""
        if not self._serial:
            try:
                with async_timeout.timeout(10):
                    power_on = await self.get_property(POWER, get_timeout(POWER))
                    if power_on == EPSON_CODES[POWER]:
                        reader, writer = await asyncio.open_connection(
                            host=self._host,
                            port=TCP_SERIAL_PORT,
                        )
                        _LOGGER.debug("Asking for serial number.")
                        writer.write(SERIAL_BYTE)
                        response = await reader.read(32)
                        self._serial = response[24:].decode()
                        writer.close()
                    else:
                        _LOGGER.error("Is projector turned on?")
            except asyncio.TimeoutError:
                _LOGGER.error(
                    "Timeout error receiving SERIAL of projector. Is projector turned on?"
                )
        return self._serial



"""Python library to control Epson projector."""

from epson_projector.error import ProjectorError, ProjectorUnavailableError

from epson_projector.projector import Projector

from epson_projector.version import __version__



"""Serial USB/RS232 connection of Epson projector module."""
import logging

import asyncio
import serial_asyncio
from serial.serialutil import SerialException
from .const import ESCVP_HELLO_COMMAND, COLON, CR, GET_CR, BUSY, ERROR, SNO
import async_timeout

_LOGGER = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10
MAX_TIMEOUTS = 3


class ProjectorSerial:
    """
    Epson Serial connector
    """

    def __init__(self, host):
        """
        Epson Serial connector

        :param str host:    Device to connect to.
        """
        self._host = host
        self._reader = None
        self._writer = None
        self._timeouts = 0
        self._isOpen = False
        self._loop = asyncio.get_running_loop()
        self._serial = None

    async def async_init(self):
        """Async init to open serial connection with projector."""

        _LOGGER.debug("Establishing serial connection")
        if self._writer and self._writer.is_closing():
            try:
                await self._writer.wait_closed()
            except:
                pass
        try:
            with async_timeout.timeout(DEFAULT_TIMEOUT):
                (
                    self._reader,
                    self._writer,
                ) = await serial_asyncio.open_serial_connection(
                    url=self._host, baudrate=9600, loop=self._loop
                )
                if self._reader and self._writer:
                    self._isOpen = True
                    self._writer.write(ESCVP_HELLO_COMMAND.encode())
                    response = await self._reader.readuntil(COLON.encode())
                    if str(response.decode().strip(CR)) == ":":
                        _LOGGER.info("Connection open")
                        return True
                    else:
                        _LOGGER.warn(
                            "Connection established, " "but wrong response %r.",
                            response,
                        )
        except asyncio.TimeoutError:
            _LOGGER.error("Timeout error during connection")
        except SerialException as se:
            _LOGGER.error(f"Problem opening serial connection: {se}")
            self._isOpen = False
        return self.closed_connection_info()

    def closed_connection_info(self):
        _LOGGER.error("Cannot open serial to Epson")
        return False

    def close(self):
        if self._writer and not self._writer.is_closing():
            _LOGGER.debug("Closing serial connection")
            self._writer.close()
            self._writer = None
            self._isOpen = False
            self._timeouts = 0

    def _check_timeout_reconnect(self):
        if self._timeouts >= MAX_TIMEOUTS:
            self._writer.close()

    async def get_property(self, command, timeout):
        """Get property state from device."""
        response = await self.send_request(timeout=timeout, command=command + GET_CR)
        if not response:
            return False
        try:
            return response.split("=")[1]
        except KeyError:
            return BUSY
        except IndexError:
            _LOGGER.error("Bad response %s", response)
            return False

    async def send_command(self, command, timeout):
        """Send command to Epson."""
        response = await self.send_request(timeout=timeout, command=command + CR)
        return response

    async def send_request(self, timeout, command):
        """Send request to Epson over serial."""
        if self._writer and not self._isOpen:
            self._writer.close()
        if self._writer is None or self._writer.is_closing():
            await self.async_init()
        if self._writer and self._isOpen and command:
            try:
                with async_timeout.timeout(timeout):
                    _LOGGER.debug("Sent to Epson: %r with timeout %d", command, timeout)
                    self._writer.write(command.encode())
                    response = await self._reader.readuntil(COLON.encode())
                    response = response[:-1].decode().rstrip(CR)
                    _LOGGER.debug("Response from Epson %r", response)
                    if response == ERROR:
                        _LOGGER.error("Error response to request %r", command)
                    else:
                        return response
            except asyncio.TimeoutError:
                _LOGGER.error("Timeout error during sending request %r", command)
                self._timeouts += 1
                self._check_timeout_reconnect()
            except SerialException as se:
                _LOGGER.error(f"Error during serial write/read: {se}")
                self.close()

        return False

    async def get_serial(self):
        """Send request for serial to Epson."""
        if not self._serial:
            response = await self.get_property(SNO, timeout=DEFAULT_TIMEOUT)
            if not response or response == BUSY:
                _LOGGER.error("Error retrieving serial number from projector")
            else:
                self._serial = response
        return self._serial



"""Const helpers of Epson projector module."""

HTTP = "http"
TCP = "tcp"
SERIAL = "serial"
HTTP_OK = 200

TCP_PORT = 3629
TCP_SERIAL_PORT = 3620
HTTP_PORT = 80
EEMP0100 = "45454d5030313030"
SERIAL_COMMAND = "0000000002000000"
SERIAL_BYTE = bytearray.fromhex(f"{EEMP0100}{SERIAL_COMMAND}")

ACCEPT_ENCODING = "gzip, deflate"
ACCEPT_HEADER = "application/json, text/javascript"
JSON_QUERY = "json_query"
DIRECT_SEND = "directsend"

ESCVPNET_HELLO_COMMAND = "ESC/VP.net\x10\x03\x00\x00\x00\x00"
ESCVPNETNAME = "ESC/VP.net"
ESCVPNAME = "ESC/VP"
ERROR = "ERR"
PWR_OFF_STATE = "04"
ESCVP_HELLO_COMMAND = "\r"
COLON = ":"
CR = "\r"
CR_COLON = CR + COLON
GET_CR = "?" + CR

POWER = "PWR"
CMODE = "CMODE"
SOURCE = "SOURCE"
VOLUME = "VOLUME"
MUTE = "MUTE"
VOL_UP = "VOL_UP"
VOL_DOWN = "VOL_DOWN"
PLAY = "PLAY"
PAUSE = "PAUSE"
FAST = "FAST"
BACK = "BACK"
TURN_ON = "PWR ON"
PWR_ON = "PWR ON"
TURN_OFF = "PWR OFF"
PWR_OFF = "PWR OFF"
ALL = "ALL"
IMGPROC_FINE = "IMGPROC_FINE"
IMGPROC_FAST = "IMGPROC_FAST"
LUMINANCE = "LUMINANCE"
MEMORY_1 = "MEMORY_1"
MEMORY_2 = "MEMORY_2"
MEMORY_3 = "MEMORY_3"
MEMORY_4 = "MEMORY_4"
MEMORY_5 = "MEMORY_5"
MEMORY_6 = "MEMORY_6"
MEMORY_7 = "MEMORY_7"
MEMORY_8 = "MEMORY_8"
MEMORY_9 = "MEMORY_9"
MEMORY_10 = "MEMORY_10"
SNO = "SNO"
BUSY = 2
BUSY_CODES = [2, "2", "02"]

EPSON_CODES = {"PWR": "01"}

EPSON_KEY_COMMANDS = {
    "PWR ON": [("KEY", "3B")],
    "PWR OFF": [("KEY", "3B"), ("KEY", "3B")],
    "HDMILINK": [("jsoncallback", "HDMILINK?")],
    "PWR": [("jsoncallback", "PWR?")],
    "SOURCE": [("jsoncallback", "SOURCE?")],
    "CMODE": [("jsoncallback", "CMODE?")],
    "VOLUME": [("jsoncallback", "VOL?")],
    "CMODE_AUTO": [("CMODE", "00")],
    "CMODE_CINEMA": [("CMODE", "15")],
    "CMODE_NATURAL": [("CMODE", "07")],
    "CMODE_BRIGHT": [("CMODE", "0C")],
    "CMODE_DYNAMIC": [("CMODE", "06")],
    "CMODE_3DDYNAMIC": [("CMODE", "18")],
    "CMODE_3DCINEMA": [("CMODE", "17")],
    "CMODE_3DTHX": [("CMODE", "19")],
    "CMODE_BWCINEMA": [("CMODE", "20")],
    "CMODE_ARGB": [("CMODE", "21")],
    "CMODE_DCINEMA": [("CMODE", "22")],
    "CMODE_THX": [("CMODE", "13")],
    "CMODE_GAME": [("CMODE", "0D")],
    "CMODE_STAGE": [("CMODE", "16")],
    "CMODE_AUTOCOLOR": [("CMODE", "C1")],
    "CMODE_XV": [("CMODE", "0B")],
    "CMODE_THEATRE": [("CMODE", "05")],
    "CMODE_THEATREBLACK": [("CMODE", "09")],
    "CMODE_THEATREBLACK2": [("CMODE", "0A")],
    "VOL_UP": [("KEY", "56")],
    "VOL_DOWN": [("KEY", "57")],
    "MUTE": [("KEY", "D8")],
    "HDMI1": [("KEY", "4D")],
    "HDMI2": [("KEY", "40")],
    "PC": [("KEY", "44")],
    "VIDEO": [("KEY", "46")],
    "USB": [("KEY", "85")],
    "LAN": [("KEY", "53")],
    "WFD": [("KEY", "56")],
    "PLAY": [("KEY", "D1")],
    "PAUSE": [("KEY", "D3")],
    "STOP": [("KEY", "D2")],
    "BACK": [("KEY", "D4")],
    "FAST": [("KEY", "D5")],
    "IMGPROC_FINE": [("IMGPROC", "01")],
    "IMGPROC_FAST": [("IMGPROC", "02")],
    "LUMINANCE_ECO": [("LUMINANCE", "01")],
    "LUMINANCE_NORMAL": [("LUMINANCE", "00")],
    "MEMORY_1": [("POPMEM", "02 01")],
    "MEMORY_2": [("POPMEM", "02 02")],
    "MEMORY_3": [("POPMEM", "02 03")],
    "MEMORY_4": [("POPMEM", "02 04")],
    "MEMORY_5": [("POPMEM", "02 05")],
    "MEMORY_6": [("POPMEM", "02 06")],
    "MEMORY_7": [("POPMEM", "02 07")],
    "MEMORY_8": [("POPMEM", "02 08")],
    "MEMORY_9": [("POPMEM", "02 09")],
    "MEMORY_10": [("POPMEM", "02 0A")],
    "LENSMEMORY_1": [("POPLP", "01")],
    "LENSMEMORY_2": [("POPLP", "02")],
    "LENSMEMORY_3": [("POPLP", "03")],
    "LENSMEMORY_4": [("POPLP", "04")],
    "LENSMEMORY_5": [("POPLP", "05")],
    "LENSMEMORY_6": [("POPLP", "06")],
    "LENSMEMORY_7": [("POPLP", "07")],
    "LENSMEMORY_8": [("POPLP", "08")],
    "LENSMEMORY_9": [("POPLP", "09")],
    "LENSMEMORY_10": [("POPLP", "0A")],
}

DEFAULT_TIMEOUT_TIME = 3
TIMEOUT_TIMES = {"PWR ON": 40, "PWR OFF": 10, "SOURCE": 5}

DEFAULT_SOURCES = {
    "HDMI1": "HDMI1",
    "HDMI2": "HDMI2",
    "PC": "PC",
    "VIDEO": "VIDEO",
    "USB": "USB",
    "LAN": "LAN",
    "WFD": "WiFi Direct",
}

SOURCE_LIST = {
    "30": "HDMI1",
    "10": "PC",
    "40": "VIDEO",
    "52": "USB",
    "53": "LAN",
    "56": "WDF",
    "A0": "HDMI2",
    "41": "VIDEO",
}

INV_SOURCES = {v: k for k, v in DEFAULT_SOURCES.items()}

CMODE_LIST = {
    "00": "Auto",
    "15": "Cinema",
    "07": "Natural",
    "0C": "Bright Cinema",
    "06": "Dynamic",
    "17": "3D Cinema",
    "18": "3D Dynamic",
    "19": "3D THX",
    "20": "B&W Cinema",
    "21": "Adobe RGB",
    "22": "Digital Cinema",
    "13": "THX",
    "0D": "Game",
    "16": "Stage",
    "C1": "AutoColor",
    "0B": "x.v. color",
    "05": "Theatre",
    "09": "Theatre Black 1/HD",
    "0A": "Theatre Black 2/Silver Screen",
}

CMODE_LIST_SET = {
    "cinema": "CMODE_CINEMA",
    "Cinema": "CMODE_CINEMA",
    "natural": "CMODE_NATURAL",
    "Natural": "CMODE_NATURAL",
    "bright cinema": "CMODE_BRIGHT",
    "Bright Cinema": "CMODE_BRIGHT",
    "dynamic": "CMODE_DYNAMIC",
    "Dynamic": "CMODE_DYNAMIC",
    "3ddynamic": "CMODE_3DDYNAMIC",
    "3D Dynamic": "CMODE_3DDYNAMIC",
    "3dcinema": "CMODE_3DCINEMA",
    "3D Cinema": "CMODE_3DCINEMA",
    "auto": "CMODE_AUTO",
    "Auto": "CMODE_AUTO",
    "3dthx": "CMODE_3DTHX",
    "3D THX": "CMODE_3DTHX",
    "bwcinema": "CMODE_BWCINEMA",
    "B&W Cinema": "CMODE_BWCINEMA",
    "adobe rgb": "CMODE_ARGB",
    "Adobe RGB": "CMODE_ARGB",
    "digital cinema": "CMODE_DCINEMA",
    "Digital Cinema": "CMODE_DCINEMA",
    "thx": "CMODE_THX",
    "THX": "CMODE_THX",
    "game": "CMODE_GAME",
    "Game": "CMODE_GAME",
    "stage": "CMODE_STAGE",
    "Stage": "CMODE_STAGE",
    "autocolor": "CMODE_AUTOCOLOR",
    "AutoColor": "CMODE_AUTOCOLOR",
    "xv": "CMODE_XV",
    "x.v. color": "CMODE_XV",
    "theatre": "CMODE_THEATRE",
    "Theatre": "CMODE_THEATRE",
    "theatre black": "CMODE_THEATREBLACK",
    "theatre black 2": "CMODE_THEATREBLACK2",
}


STATE_UNAVAILABLE = "unavailable"
